%   File   : /usr/lib/prolog/read_sent
%   Author : R.A.O'Keefe
%   Updated: 21 May 84
%   Purpose: to provide a flexible input facility
%   Needs  : memberchk from lib(sets) or lib(utils).

/*
 rcsid('$Id: slereader.pl,v 1.5 1992/04/16 13:43:35 pleuk Exp $').
 $log$
File:	/home/user2/jo/Pleuk/Notes/read_sent
Date:	Wed May 29 16:15:37 1991
Adapted by:	Jonathan Calder


Some heavy adaptations, in particular to allow more sophisticated
handling of alternate terminators. Our strategy is to pass around as a
list all characters read from the current definition, so that we can
have more control over terminators, escape characters, etc.  This
means that efficiency takes a bit of a tumble.  To get around this,
some partial evaulation has been done so that we can get more through
pattern matching.  In particular, the multiline and single line
comment characters have been hard-wired. 

In order for correct interaction with Pleuk, we have to make the
assumption that the last item in every file comes with a terminator,
and that the end_of_file marker will *not* be read in conjunction with
some definition.  Trying to work round this restriction, for example
by enumerating first the definition then an end of file flag is going
to be a real pain. 


*/

*/


:- public
	case_shift/2,
	chars_to_words/2,
	is_digit/1,
	is_endfile/1,
	is_layout/1,
	is_letter/1,
	is_lower/1,
	is_newline/1,
	is_paren/2,
	is_period/1,
	is_punct/1,
	is_upper/1,
	read_line/1,
	read_sent/1,
	read_until/2,
	trim_blanks/2.



%   read_until(Delimiters, Answer)
%   reads characters from the current input until  a  character  in  the
%   Delimiters  string  is  read.  The characters are accumulated in the
%   Answer string, and include the closing  delimiter.   Prolog  returns
%   end-of-file as ^Z (26)  regardless of the user's assignment (e.g. if
%   you use ^D as end of file, Prolog still returns ^Z).  The end of the
%   file is always a delimiter.

read_until(Stream, Delimiters, Escape, String) :-
	eccs_sys_get0(Stream, Char),
	read_until(Stream, [Char], Delimiters, Escape, String).


read_until(Stream, [Char|Chars], Delimiters, Escape, [Char|Chars]) :-
    	eccs_terminated([Char|Chars], Delimiters), !.
read_until(Stream, [Char|Chars], Delimiters, Escape, [-1|Chars]) :-
	eccs_possible_eof(Char), !.
read_until(Stream, [42, 47|Cs], Delimiters, Escape, String) :- % /* -> space
    	!,
    	read_until_end_long_comment( Stream, "/*"), !,
	read_until(Stream, [32|Cs], Delimiters, Escape, String).
read_until(Stream, [37|Cs], Delimiters, Escape, String) :-
    	!,
    	read_until_end_short_comment(Stream), !,
	read_until(Stream, [32|Cs], Delimiters, Escape, String).
read_until(Stream, Cs, Delimiters, String) :-
	eccs_sys_get0(Stream, Char),
	read_until(Stream, [Char|Cs], Delimiters, Escape, String).


eccs_terminated([Char, Term|Chars], Delimiters) :-
    is_layout(Char), eccs_memberchk(Term, Delimiters).


read_until_end_long_comment(Stream, [Char2, Char1]) :-
	eccs_sys_get0(Stream, C),
	read_until_end_long_comment(Stream, C, Char1, Char2).

read_until_end_long_comment(Stream, C, _Char1, _Char2) :-
	eccs_possible_eof(C), !,
	eccs_error([end, of, file, in, comment]).
read_until_end_long_comment(Stream, C, C, Char2) :-
    	eccs_sys_get0(Stream, Char),
	(Char = Char2 -> !;
	 read_until_end_long_comment(Stream, Char, C, Char2)).
read_until_end_long_comment(Stream, _, Char1, Char2) :-
    	eccs_sys_get0(Stream, Char),
	read_until_end_long_comment(Stream, Char, Char1, Char2).

read_until_end_short_comment(Stream) :-
    	eccs_sys_repeat,
	eccs_sys_get0(Stream, Char),
	is_newline(Char), !.

%   The following predicates define useful character classes.

is_newline(10).				% the line terminator linefeed
is_newline(13).

is_endfile(-1).				% the file terminator and friends
is_endfile(26).
is_endfile(4).

is_layout(Char) :-
	eccs_sys_lte(Char, 0' ).	% includes tab, newline, ^S, &c

is_lower(Char) :-
	eccs_sys_lte(0'a, Char), eccs_sys_lte(Char, 0'z). % lower case letter

is_upper(Char) :-
	eccs_sys_lte(0'A, Char), eccs_sys_lte(Char, 0'Z). % upper case letter

is_letter(Char) :-
	is_lower(Char) ; is_upper(Char).

is_digit(Char) :-
	eccs_sys_lte(0'0, Char), eccs_sys_lte(Char, 0'9). % decimal digit

is_period(Char) :-
	eccs_memberchk(Char, ".!?").		% sentence terminator

is_punct(Char) :-
	eccs_memberchk(Char, ",;:").		% other punctuation mark

is_paren(Left,Right) :-				% brackets
	eccs_memberchk([Left,Right], ["()","[]","{}","<>"]).


%   trim_blanks(RawInput, Cleaned)
%   removes leading and trailing layout characters  from  RawInput,  and
%   replaces  internal  groups  of  layout  characters by single spaces.
%   Thus trim_blanks(<|TAB TAB a SP ^M ^E b ^Z|>, "a b") would be true.

trim_blanks([Char|Chars], Cleaned) :-
	is_layout(Char), !,
	trim_blanks(Chars, Cleaned).
trim_blanks([Char|Chars], [Char|Cleaned]) :- !,
	trim_blanks_rest_word(Chars, Cleaned).
trim_blanks([], []).


trim_blanks_rest_word([Char|Chars], Cleaned) :-
	is_layout(Char), !,
	trim_blanks_next_word(Chars, Cleaned).
trim_blanks_rest_word([Char|Chars], [Char|Cleaned]) :- !,
	trim_blanks_rest_word(Chars, Cleaned).
trim_blanks_rest_word([], []).


trim_blanks_next_word([Char|Chars], Cleaned) :-
	is_layout(Char), !,
	trim_blanks_next_word(Chars, Cleaned).
trim_blanks_next_word([Char|Chars], [32,Char|Cleaned]) :- !,
	trim_blanks_rest_word(Chars, Cleaned).
trim_blanks_next_word([], []).



%   chars_to_words(Chars, Words)
%   parses a list of characters (read by read_until) into a list of
%   tokens, where a token is
%	'X' for X a period or other punctuation mark, e.g. ';'
%	atom(X) for X a sequence of letters, e.g. atom(the)
%	integer(X) for X a sequence of digits, e.g. integer(12)
%	apost for '
%	aposts for 's
%	string(X) for X "..sequence of any.."
%   Thus the string "the "Z-80" is on card 12." would be parsed as
%	[atom(the),string('Z-80'),atom(is),atom(on),atom(card),
%	 integer(12),'.'].
%   It is up to the sentence parser to decide what to do with these.
%   Note that the final period, if any, is retained.  The parser may
%   need it.

chars_to_words(Chars, Words) :-
	chars_to_words(Words, Chars, []).


chars_to_words([Word|Words]) -->
	chars_to_word(Word), !,
	chars_to_words(Words).
chars_to_words([]) --> [].


chars_to_word(Word) -->
	[Char], {is_layout(Char)}, !,
	chars_to_word(Word).
chars_to_word(atom(Word)) -->
	[Char], {is_letter(Char)}, !,
	chars_to_atom(Chars),
	{case_shift([Char|Chars], Name)},
	{name(Word, Name)}.
chars_to_word(integer(Word)) -->
	[Char], {is_digit(Char)}, !,
	{plus(Init, 48, Char)},		% integer value of char
	chars_to_integer(Init, Word).
chars_to_word(aposts) -->
	"'s", !.
chars_to_word(apost) -->
	"'", !.
chars_to_word(string(Word)) -->
	"""", !,			% 34 is """"
	chars_to_string(0'", String),
	{name(Word, String)}.
chars_to_word(Punct) -->
	[Char],
	{name(Punct, [Char])}.


%   chars_to_atom(Tail)
%   reads the remaining characters of a word.  Case conversion  is  left
%   to  another  routine.   In this application, a word may only contain
%   letters but they may be in either case.  If you want to parse French
%   you will have to decide what to do about accents.  I suggest putting
%   them after the vowel, and adding a clause
%	chars_to_atom([Vowel,Accent|Chars]) -->
%		[Vowel],	{accentable_vowel(Vowel)},
%		[Accent],	{accent_for(Vowel, Accent)},
%		!.
%	with the obvious definitions of accentable_vowel and accent_for.
%   Note that the Ascii characters ' ` ^ are officially  designated  the
%   "accent acute", "accent grave", and "circumflex".  But this file was
%   originally written for an English parser and there was no problem.

chars_to_atom([Char|Chars]) -->
	[Char], {is_letter(Char)}, !,
	chars_to_atom(Chars).
chars_to_atom([]) --> [].


%   case_shift(Mixed, Lower)
%   converts all the upper case letters in Mixed to lower  case.   Other
%   characters (not necessarily letters!) are left alone.  If you decide
%   to accept other characters in words only chars_to_atom has to alter.

case_shift([Upper|Mixed], [Letter|Lower]) :-
	is_upper(Upper), !,
	plus(Upper, 32, Letter),	%  Letter is Upper+32
	case_shift(Mixed, Lower).
case_shift([Other|Mixed], [Other|Lower]) :-
	case_shift(Mixed, Lower).
case_shift([], []).


%   chars_to_integer(Init, Final)
%   reads the remaining characters of an integer which starts  as  Init.
%   NB:  this  parser  does  not  know about negative numbers or radices
%   other than 10, as it was written for PDP-11 Prolog.

chars_to_integer(Init, Final) -->
	[Char], {is_digit(Char)}, !,
	{Next is Init*10-"0"+Char},
	chars_to_integer(Next, Final).
chars_to_integer(Final, Final) --> [].


%   chars_to_string(Quote, String)
%   reads the rest of a string which was opened by  a  Quote  character.
%   The  string is expected to end with a Quote as well.  If there isn't
%   a matching Quote, the attempt to parse the  string  will  FAIL,  and
%   thus the whole parse will FAIL.  I would prefer to give some sort of
%   error  message and try to recover but that is application dependent.
%   Two adjacent Quotes are taken as one, as they are in Prolog itself.

chars_to_string(Quote, [Quote|String]) -->
	[Quote,Quote], !,
	chars_to_string(Quote, String).
chars_to_string(Quote, []) -->
	[Quote], !.
chars_to_string(Quote, [Char|String]) -->
	[Char], !,
	chars_to_string(Quote, String).


%   read_line(Chars)
%   reads characters up to the next newline or the end of the file, and
%   returns them in a list, including the newline or end of file.  When
%   you want multiple spaces crushed out, and the newline dropped, that
%   is most of the time, call trim_blanks on the result.

read_line(Stream, Chars) :-
	is_newline(NL),
	read_until(Stream, [NL], Chars).


%   read_sent(Words)
%   reads characters up to the next period, which may be  several  lines
%   distant from the start, skips to the end of that line, and turns the
%   result  into  a  list of tokens.  It can happen that the sentence is
%   not well formed, if say there is an unmatched double quote.  In that
%   case all the characters will still be read, but chars_to_words  will
%   fail  and  so  read_sent  will fail.  read_sent will NOT try to read
%   another sentence.

read_sent(Stream, Words) :-
	read_until(Stream, "!?.", Chars),
	is_newline(NL),
	read_until(Stream, [NL], _),		% skip to end of line
	!,
	chars_to_words(Chars, Words),
	!.




/*

Generic reader, following prolog convention that an item in a file is
bounded by an end of item character and a layout character (including
end of file).

We assume that we always have the last two characters from the file
available for scrutiny.  This means that we start this routine with
two spaces to fake this situation. Any tokenizers have therefore to
make sure that they allow whitespace at the beginning of definitions. 

*/


eccs_possible_eof(-1).
eccs_possible_eof(26).
eccs_possible_eof(4).



/*

read_item(Item)  

get next item from file

*/


:- dynamic eccs_peeked_char/1.

eccs_peek(Stream, N, 



eccs_peek_read(
